// Functions to support WCTasks related WCR GUI.
// author: JANH


// element used to measure string length in pixels of "Add predecessors" menu/list 
document.write('<span id="dropmenudiv_ruler" style="visibility:hidden; white-space:nowrap; position:absolute;"></span>');

// language strings
var langStrTaskName = '?translation missing?';
var langStrAssignedTo = '?translation missing?';
var langStrDueDate = '?translation missing?';
var langStrNoTaskToAdd = '?translation missing?';
var langStrNoPredecessor = '?translation missing?';
var langStrStartDateReset = '?translation missing?';
var langStrNobodyOption = '?translation missing?';

var langStrInvalidLeadTime = '?translation missing?';
var langStrNoDueDateErr = '?translation missing?';
var langStrDueDateInPastErr = '?translation missing?';
var langStrStartDateInPastErr = '?translation missing?';
var langStrTaskSequenceErr = '?translation missing?';
var langStrNonnegativeValueErr = '?translation missing?';

var langStrDeletePredecessor = '?translation missing?';

var tasksMenu = new Array();

var taskMenuWidth = 300;
var predTableContainerID = null;

// array of task predecessor IDs (auxiliary array) 
var predecessorIDs;

// array of task successor IDs (auxiliary array) 
var successorIDs;

/*
 * structure of tasks array element:
 * {id:string, predecessor:boolean, name:string, assignedTo:string (name), dueDate:int (timestamp),
 *  taskTypeID:string, 
 *  successors:{id:string, idx:int} //successors' IDs and corresponding indices in 'tasks' array
 * }
 */
var tasks;

/*
 * structure of taskTypes array element:
 * {id:string, name:string, creatorName:string, days:int, hours:int, 
 *	statuses:{id:string, name:string, taskTypeId:string}
 * }
 */
var taskTypes;

/*
 * structure of "translation table" predecessorIdxs array element (used in projdetailswctasks.xsl):
 * {id:string, value:int}
 */
var predecessorIdxs;

var userTimezoneOffset = 0;

var dueDate = null;
var dueDateChangedEventsEnabled = true;
var startDate = null;
var leadDays = 0;
var leadHours = 0;

var actTaskType = null;
var actTaskStatus = null;

var currentTaskID = null;
var taskStatusID = null;
var taskNameChangedManually = false;
var loggedInUserID = null;
var taskOwnerID = null;
var linkPredecsDocs = false;
var allowManualStart = false;
var startTaskOptionIdx = null;
var newProjectStatusID = null;
var origStatusName = '';

// wctask state info
var newTaskComposing = true;
var isTemplate = false;
var isEditable = false; // true only if user has permission to change the task and the task is not completed
var isStatusEditable = false;

var dueDateRelatedChange = false;


// initialization method that needs to be called prior to first usage of
// this javascript source code
// Called from projwctasktemplates.xsl
function initWCTasksJS(elementId, tasksData, taskTypesData, editable, statusEditable) {
	predTableContainerID = elementId;
	tasks = tasksData;
	taskTypes = taskTypesData;
	isEditable = editable;
	isStatusEditable = statusEditable;

	if (isNaN(userTimezoneOffset)) userTimezoneOffset = 0;

	// find out which tasks are already predecessors of current task
	for (var i = 0; i < predecessorIDs.length; i++) {
		var idx = getIndexFromId(tasks, predecessorIDs[i]);
		if (idx != null) tasks[idx].predecessor = true;
	}

	// get the indices of individual successors (tasks) in 'tasks' array
	for (var t = 0; t < tasks.length; t++) {
		var succList = tasks[t].successors;
		if (!succList) continue;
		for (var i = 0; i < succList.length; i++) {
			succList[i].idx = getIndexFromId(tasks, succList[i].id);
			if (succList[i].idx == null) alert('successor not found in project task list!');
		}
	}

	// initialize selected task type
	changeActTaskType(getIndexFromId(taskTypes, getElement('taskType').value));

	// initialize actTaskStatus value
	if (!taskStatusID || (taskStatusID.length == 0)) taskStatusID = '00001_0000000001';
	if (actTaskType != null) {
		actTaskStatus = getIndexFromId(taskTypes[actTaskType].statuses, taskStatusID);
	}

	// task owner selector init
	if (getElement('taskOwner') != null) {
		if (taskOwnerID && (taskOwnerID.length > 0)) {
			if ((taskOwnerID == null) || (taskOwnerID.length == 0)) getElement('taskOwner')[0].selected = true;
			else if (taskOwnerID == 'U' + loggedInUserID) selectItemOfSelectInput('taskOwner', 'current_user'); // current user
			else selectItemOfSelectInput('taskOwner', taskOwnerID);
		}
	}

	// set lead time and due date values differently when the form is editable and when it's not
	if (!isEditable) {
		if (isNaN(leadDays)) {
			// set leadDays and leadHours from task type if not specified in given task (true for very old tasks)
			leadDays = taskTypes[actTaskType].days;
			leadHours = taskTypes[actTaskType].hours;
		}
	} else {
		// lead time variables init
		if (actTaskType == null) {
			leadDays = leadHours = 0;
		} else {
			leadDays = parseInt(getElement('lengthInput').value);
			leadHours = parseInt(getElement('lengthTime').value);
			var taskType = taskTypes[actTaskType];
			if (isNaN(leadDays)) leadDays = taskType.days;
			if (isNaN(leadHours)) leadHours = taskType.hours;
		}
	}

	// due date Date object still needs to be made from current dueDate value
	if (dueDate && !isNaN(dueDate)) {
		dueDate = new Date(dueDate);
	}

	// handle start task options
	var optionElems = document.getElementsByName('startTaskOption');
	if (optionElems && (optionElems.length > 0)) {
		for (var i = 0; i < optionElems.length; i++) {
			if (optionElems[i].checked) startTaskOptionIdx = i;
		}
		// default value for startTaskOptionIdx is "Automatically at ..."
		if (startTaskOptionIdx == null) startTaskOptionIdx = 1;
		optionElems[startTaskOptionIdx].checked = true;
	}

	// initialize "Automatically link documents from predecessor tasks when task is started" check box
	var linkPredecsDocsElem = getElement('linkPredecsDocsWhenStarted');
	if (linkPredecsDocsElem) linkPredecsDocs = linkPredecsDocsElem.checked;
	showLinkPredecsDocsCheckbox(existsPredecessor());

	// initialize "can start task manually check box"
	var manualStartElem = getElement('canStartTaskManually');
	if (startTaskOptionIdx == 0) {
		manualStartElem.checked = true;
		manualStartElem.disabled = true;
	}
	allowManualStart = manualStartElem.checked;

	var projStatusCheckbox = getElement('projectStatusCheckbox');
	var projStatSelectElem = getElement('projectStatusSelect');
	if (projStatSelectElem) projStatSelectElem.disabled = !projStatusCheckbox.checked;

	// the new project status is set by default to 'completed'
	if (!newProjectStatusID || (newProjectStatusID.length == 0)) newProjectStatusID = '00001_0000000003';

	selectItemOfSelectInput('projectStatusSelect', newProjectStatusID);

	// generate GUI
	updateWCTaskGenInfoPage(false);	

	// task status selector init
	selectItemOfSelectInput('taskStatusInput', taskStatusID);
	getElement('taskStatusID').value = taskStatusID;

	dueDateRelatedChange = false;
}

/**
 * Execute a callback as soon as the due date picker is available in the DOM
 * @param callback
 * @returns
 */
function callWhenDueDatePickerIsInDOM(callback){
	var div = $("#relativedatetimepicker_duedate");
	var picker = div.data("relativeDateTimePicker");
	if(picker){
		// The picker is in the DOM, execute the callback
		if(callback){
			callback(picker);
		}
	} else {
		// The picker is not available yet, create a promise to wait for it. Pass the resolve function to the DOM, so it can be resolved once the picker is available (see projwctasktemplates.xsl).
		var pickerPromise = div.data("pickerPromise");
		if(!pickerPromise){
			pickerPromise = new Promise(function(resolve){
				div.data("pickerPromiseResolver", resolve);
			});
			div.data("pickerPromise", pickerPromise);
		} 
		pickerPromise.then(function(picker){
			if(callback){
				callback(picker);
			}
		});
	}
}

/**
 * Displays a relative date time picker in readonly mode (only text) in the DOM (id used in the 'start date' options)
 * @param {JQueryObj} element - Element of where to put the picker
 * @param {Object} pickerValue - See RelativeDateTimePicker -> setValue 
 */
function displayReadonlyPicker(element, pickerValue) {
	require(["component/relativedatetimepicker/RelativeDateTimePicker"], function (RelativeDateTimePicker) {
		var picker = new RelativeDateTimePicker({
			"supportEdit": false
		});
		picker.setValue(pickerValue);
		var dateElement = picker.getElement().css({
			"display": "inline-block",
			"margin": "0px 6px"
		});
		element.empty().html(dateElement);
	});
}

/**
 * Initialize the due date picker: register the change event handler
 */
function initDueDatePicker() {
	var picker = $("#relativedatetimepicker_duedate").data("relativeDateTimePicker");
	if (picker) {
		picker.on("change", function () {
			var value = picker.getValue();
			dueDateChanged(value);
		});
	} else {
		console.log("Due date picker not found. Could not initialise the due date picker.");
	}
}

/**
 * the changed event handler
 * @param {object} value - The new value of the relative date time picker, for data format see RelativeDateTimePicker -> getValue
 */
function dueDateChanged(value) {
	if (value && value.time > 0) {
		dueDate = new Date(value.time);
	} else {
		dueDate = null;
	}
	updateStartDateGUI();
	dueDateRelatedChange = true;
}

/**
 * Get the due date from the resolvable relativedatetimepicker
 * @returns {object} 
 */
function getDueDatePickerValue() {
	var picker = $("#relativedatetimepicker_duedate").data("relativeDateTimePicker");
	if (picker) {
		return picker.getValue();
	} else {
		console.log("Due date picker not found. Could not get due date value.");
		return null;
	}
}

/**
 * Set the due date value of the relativedatetimepicker
 * @param {Object} value (see RelativeDateTimePicker)
 */
function setDueDatePickerValue(value) {
	var picker = $("#relativedatetimepicker_duedate").data("relativeDateTimePicker");
	if (picker) {
		picker.setValue(value);
	}else {
		console.log("Due date picker not found. The value of the due date picker could not be set.");
	}
}

// workaround for problems with selecting items of select HTML input
function selectItemOfSelectInput(elemId, itemVal) {
	var selElems = getElement(elemId);
	if ((typeof selElems == 'undefined') || !selElems) return false;
	for (var i = 0; i < selElems.length; i++) {
		if (selElems[i].value == itemVal) {
			selElems[i].selected = true;
			return true;
		}
	}
	return false;
}

// method using ruler to measure length of given string.
// the returned length is in pixels
function measureStringLength(str) {
	var ruler = getElement('dropmenudiv_ruler');
	ruler.innerHTML = str;
	return ruler.offsetWidth;
}

// trims given string (rendered using given font) to given length in pixels
function trimStringToPx(str, px) {
	var trimmed = str;
	while (measureStringLength(trimmed) > px - 5) { //('-5' is a safety gab)
		trimmed = trimmed.substring(0, trimmed.length - 1);
	}
	if (trimmed.length < str.length) trimmed = trimmed.substring(0, trimmed.length - 3) + '...';
	return trimmed;
}


/*************************************************************************/


function getIndexFromId(data, id) {
	for (var i = 0; i < data.length; i++) {
		if (data[i].id == id) return i;
	}
	return null;
}

function getIndexFromName(data, name) {
	for (var i = 0; i < data.length; i++) {
		if (data[i].name.toUpperCase() == name.toUpperCase()) return i;
	}
	return null;
}

function arrayContainsValue(data, val) {
	for (var i = 0; i < data.length; i++) {
		if (data[i] == val) return true;
	}
	return false;
}


function getAllSuccessorIndicesOf(masterIdx) {
	var res = new Array();
	if (masterIdx == null) return res;

	// save index of actual successor task to 'res' (=result) array
	res.push(masterIdx);

	// load successors of given task...
	var succList = tasks[masterIdx].successors;
	if (!succList || (succList.length == 0)) return res;

	// ...and find all their subsequent successors (recursion)
	for (var i = 0; i < succList.length; i++) {
		getSuccessorIndices(masterIdx, res, succList[i].idx);
	}
	return res;
}

function getSuccessorIndices(masterIdx, res, idx) {
	// save index of actual successor task to 'res' (=result) array
	res.push(idx);

	// detection of a cycle within task dependencies
	if (masterIdx == idx) {
		alert('A cycle within tasks dependencies found: ' + res.toString());
		return;
	}

	// check whether the actual task has its own successors
	var succList = tasks[idx].successors;
	if (!succList || (succList.length == 0)) return;

	// if so, call this method recursively
	for (var i = 0; i < succList.length; i++) {
		getSuccessorIndices(masterIdx, res, succList[i].idx);
	}
}


function changeActTaskType(newValue) {
	if (newValue == null) {
		actTaskStatus = null;
	} else {
		actTaskStatus = 1;
	}
	actTaskType = newValue;
}

/**
 * Called from projwctasktemplates.xsl
 * @param elem
 * @param projectid
 * @param callback
 * @returns
 */
function selectTaskType(elem, projectid, callback) {
	changeActTaskType(getIndexFromId(taskTypes, elem.value));
	require(["component/relativedatetimepicker/ResolvableRelativeDateTimePicker"], function () {
		if (actTaskType != null) {
			var taskNameElem = getElement('taskName');
			if ((taskNameElem.value == null) || (taskNameElem.value.length == 0)) taskNameChangedManually = false;
			if (!taskNameChangedManually) {
				// find a non-existing task name derived from selected task type name
				var actTaskTypeID = taskTypes[actTaskType].id;
				// make the name safe for usage in a regular expression
				var escapedName = taskTypes[actTaskType].name.replace(/([.?*+^$[\]\\(){}|-])/g, "\\$1");
				var taskNamePattern = new RegExp('^' + escapedName + '_\\d+$', 'i');
				var maxIdx = 0;
				var task;
				for (var i = 0; i < tasks.length; i++) {
					task = tasks[i];
					var taskName = task.name;
					if ((task.taskTypeID == actTaskTypeID) && taskNamePattern.test(taskName)) {
						// found a task of the same task type and name that fits the pattern of automatic task naming.
						// get the index of this automatic name (number at the end)
						var idx = parseInt(taskName.substr(taskName.lastIndexOf('_') + 1));
						if (!isNaN(idx) && (maxIdx < idx)) maxIdx = idx;
					}
				}
				// having the last index of automatically generated task create a name of current task
				var taskTypeName = taskTypes[actTaskType].name;
				maxIdx += 1;
				// verify the generated name does not exist
				while (getIndexFromName(tasks, taskTypeName + '_' + maxIdx) != null) maxIdx++;
				taskNameElem.value = taskTypeName + '_' + maxIdx;
			}

			// set the attribute in the relative date time picker
			var duedate_attribute_id = taskTypes[actTaskType].duedate_attribute_id;
			var duedate_offset = taskTypes[actTaskType].duedate_offset;

			initializeDueDatePicker(taskTypes[actTaskType], duedate_attribute_id, duedate_offset, callback);
		}
	});
}

/**
 * Set the initial value of the relative due date picker based on info from the task type, and render the page.
 * @param picker
 * @param taskType
 * @param duedate_attribute_id
 * @param duedate_offset
 * @param callback
 * @returns
 */
function initializeDueDatePicker(taskType, duedate_attribute_id, duedate_offset, callback){

	callWhenDueDatePickerIsInDOM(function(picker){
		if (duedate_attribute_id) {
			// offset mode
			if (taskType.duedate_editable === "0") {
				// do not allow breaking the link to the attribute
				picker.setDisabled(true);
			} else {
				picker.setDisabled(false);
			}
	
			// we have a relative due date to be calculated
			picker.setValue({
				item_id: duedate_attribute_id,
				offset: duedate_offset
			}).then(function (value) {
				dueDateChanged(value);
				updateWCTaskGenInfoPage(false);
			}, function (value) {
				dueDateChanged(value);
				updateWCTaskGenInfoPage(false);
			});
		} else {
			// specific date mode
			picker.setDisabled(false);
			picker.openCategorySpecificDate();
			var time = taskType.dueDate ? taskType.dueDate : getNow();
			var value = {
				time: time,
				item_id: null,
				offset: null
			};
			picker.setValue(value);
		}
		
		updateWCTaskGenInfoPage(true);
		
		// execute callback if defined
		if (callback)
			callback();
	});

}

function selectTaskStatus(elem) {
	if (actTaskType == null) return;
	actTaskStatus = getIndexFromId(taskTypes[actTaskType].statuses, elem.value);
	if (actTaskStatus == null) alert('actual task status can\'t be found');
	// store the new value of task status to hidden input of related HTML form.
	getElement('taskStatusID').value = elem.value;
}

function selectStartTaskOption(elem) {
	var optionElems = document.getElementsByName('startTaskOption');
	for (var i = 0; i < optionElems.length; i++) {
		if (optionElems[i].checked) startTaskOptionIdx = i;
	}

	var manualStartElem = getElement('canStartTaskManually');
	manualStartElem.disabled = elem.value == 0;
	if (manualStartElem.disabled) manualStartElem.checked = true;
	else manualStartElem.checked = allowManualStart;

	recalculateDueDate('StartTaskOptionChanged');
}

function selectLinkPredecsDocs(elem) {
	linkPredecsDocs = elem.checked;
}

function selectAllowManualStart(elem) {
	allowManualStart = elem.checked;
}

function selectProjectStatusCheckbox(elem) {
	var selector = getElement('projectStatusSelect');
	selector.disabled = !elem.checked;
}

function daysHoursLengthChanged(daysElem, hoursElem) {
	var days = parseInt(daysElem.value);
	var hours = parseInt(hoursElem.value);
	// handle wrong lead time values
	if (isNaN(days) || (isNaN(hours))) {
		alert(langStrInvalidLeadTime);
		if (isNaN(days)) days = 0;
		else if (isNaN(hours)) hours = 0;
	} else if ((days < 0) || (hours < 0)) {
		alert(langStrNonnegativeValueErr);
		if (days < 0) days = 0;
		else if (hours < 0) hours = 0;
	}
	// handle case when days or hours field was reset due to wrong value of user input
	daysElem.value = days;
	hoursElem.value = hours;

	if (hours > 23) {
		days += Math.floor(hours / 24);
		hours = hours % 24;
	}

	// set the new lead time values
	updateLeadTimeGUI(days, hours);

	// recalculate due date based on the original start date
	recalculateDueDate('LeadTimeChanged');
}

function setDueDate(date) {
	// processing of invalid/null date entered through UI
	try {
		if (isNaN(date.getTime())) date = null;
	} catch (err) {
		date = null;
	}

	dueDate = date;

	// update the rest of GUI according to changed due date
	updateStartDateGUI();

	// (no "recalculation" of due date is needed - simply set the new due date)
	dueDateRelatedChange = true;
}


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

function updateWCTaskGenInfoPage(taskTypeChanged) {
	// fill task statuses dropdown list
	generateTaskStatusContent('taskStatusTD');

	// set lead time values
	if (taskTypeChanged) {
		// reset user defined lead time to default value of the new task type
		updateLeadTimeGUI(null, null);

		// due date recalculation based on performed changes
		recalculateDueDate('PredecessorsChanged');

		// restrict assignee drop down visibility
		restrictAssigneeVisibility();
	} else {
		// (do not reset the lead time)
		updateLeadTimeGUI();

		// due date recalculation based on performed changes
		recalculateDueDate();
	}
}

function fetchTaskStatuses(tasktypeid) {
	//GetTaskTypes.jsp?verbose=1&loadicons=0&&tasktypeid=00002_0000000028
	try {

		ajax({
			url: "GetTaskTypes.jsp?verbose=1&loadicons=0&tasktypeid=" + tasktypeid,
			type: "GET",
			data: "xml",
			debug: false,
			synchronous: false,
			onError: function (status, objHdl, description) {
				// Don't show an alert when the task statuses fail to load.
				// Why? Because this will always happens when using a create task button
				// => it enters all the information on the page and redirects immediately to another page
				// At that time this ajax call is still loading and thus it will fail.
				console.log("[ERROR] Failed to load Task Statuses. Description: " + description);
			},
			onSuccess: function (xml) {
				var array = new Array();

				var nodeList = xml.getElementsByTagName("status_list")[0].getElementsByTagName("task_type_status_view");
				if (nodeList != null && nodeList.length > 0) {
					var node = 0;
					for (node = 0; node < nodeList.length; node++) {
						//alert("LoadDocTemplatesForProject() > AJAX response > Processing found template documents.");
						var statusnode = nodeList[node];
						var wcstatusnode = statusnode.getElementsByTagName("wc_status")[0];
						var id = WCXML.GetElementAttribute(wcstatusnode, "id");
						var name = WCXML.GetElementText(wcstatusnode, "status_name");
						var tasktypeid = WCXML.GetElementAttribute(statusnode.getElementsByTagName("wctask_type")[0], "id");

						var dict = new Array();
						dict["id"] = id;
						dict["name"] = name;
						dict["taskTypeId"] = tasktypeid;
						var jsonstring = "{id:'" + id + "',name:'" + name + "',taskTypeId:'" + tasktypeid + "'}";
						array.push(dict);
					}
				}
				taskTypes[actTaskType].statuses = array;
				generateTaskStatusContent('taskStatusTD');

			} //end: AJAX success callback					
		}); // End Ajax code
	} catch (err) {
		alert("Error: " + err);
	}

}

function generateTaskStatusContent(elementId) {
	var str = '';
	var elem = getElement(elementId);
	if (!elem) return;

	if (isEditable || isStatusEditable) {
		// if user has ability to change the task status 
		// (isEditable = false when task is completed - this is why isStatusEditable is essential here)
		if (actTaskType != null) {
			str += '<select id="taskStatusInput" name="taskStatusInput" onchange="selectTaskStatus(this)" class="form-control">';
			var statuses = taskTypes[actTaskType].statuses;
			if (statuses.length == 0) {
				fetchTaskStatuses(taskTypes[actTaskType].id);
			} else {
				for (var i = 0; i < statuses.length; i++) {
					// do not include "Completed" status in task status selector when the task is just under creation
					if (newTaskComposing && (statuses[i].id == '00001_0000000003')) continue;
					// if modifying task in template, hide all statuses except "Created"
					else if (isTemplate && (statuses[i].id != '00001_0000000001')) continue;
					str += '<option value="' + statuses[i].id + '">' + WCXML.xmlencode(statuses[i].name) + '</option>';
				}
			}
			str += '</select>';
		} else {
			str += '<select id="taskStatusInput" name="taskStatusInput" disabled="disabled" class="form-control"/>';
		}
	} else {
		// if user has no permission to edit the task - no matter if the task is completed or not
		str += '<p>' + origStatusName + '</p>';
	}
	setElementHtml(elementId, str);
}

/**
 *	Event triggered by a change of "Task Type" on Task Creation	
 *	Process state of the "Allowed Assignee" controls based on the selected Task Type
 */
function restrictAssigneeVisibility() {
	var taskMainAssigneeRowElem = $('#taskMainAssigneeRow');
	var freeMainAssigneeTipRow = $('#freeMainAssigneeTipRow');
	if (taskMainAssigneeRowElem && freeMainAssigneeTipRow && (actTaskType !== undefined && actTaskType !== null)) {
		//is the selected Task a 'restricted' one?
		var restrictAssignees = taskTypes[actTaskType].restrictAssignees;
		if (restrictAssignees === '1') {
			//remove the <NOBODY> option as we will not allow this task to be created unassigned
			$("#taskOwner option[value='nobody']").remove();

			//does the parent restricted Task Type have a pre-selected Allowed Assignee reference?
			var ttPreselectedMA_ID = taskTypes[actTaskType].preselectedMainAssigneeID;
			if (ttPreselectedMA_ID && ttPreselectedMA_ID.length > 0) {
				// show the Allowed Assignee row
				taskMainAssigneeRowElem.show();
				freeMainAssigneeTipRow.hide();

				//make the dropdown select the pre-selected MA option 
				$('#taskMainAssignee option[value="' + ttPreselectedMA_ID + '"]').prop('selected', true);
				//and disable the selector as it cannot be changed
				$('#taskMainAssignee').prop('disabled', 'disabled');
				//make sure to keep track of the preselected MA ID reference
				$('#taskMainAssigneePreselected').val(ttPreselectedMA_ID);
				//update the "Assigned to" dropdown and select the corresponding option to the current MA
				$('#taskOwner').val('G' + ttPreselectedMA_ID);
			} else {
				//Allowed Assignee will be taken from the initial assignee
				//enable the dropdown and select the first option
				$("#taskMainAssignee").val($("#taskMainAssignee option:first").val()).prop('disabled', false);
				//hide the selector row and show the free-MA-assignment tip
				taskMainAssigneeRowElem.hide();
				freeMainAssigneeTipRow.show();
			}
		} else {
			//make sure the "Assigned to" select element contains <NOBODY> option as we want to allow creating the task unassigned
			var nobodyExists = false;
			$('#taskOwner option').each(function () {
				if (this.value == 'nobody') {
					nobodyExists = true;
					return false;
				}
			});

			if (!nobodyExists) {
				$("#taskOwner").prepend("<option value='nobody'>&lt;" + langStrNobodyOption + "&gt;</option>").val('');
				//also make this option the selected one
				$('#taskOwner').val("nobody");
			}

			// not a restricted task type > hide the Allowed Assignee row altogether
			taskMainAssigneeRowElem.hide();
			freeMainAssigneeTipRow.hide();
		}
	}
}

/*************************************************************************/
/* methods related to task predecessors table */
/*************************************************************************/

function deleteTaskPredecessor(idx) {
	if (tasks[idx]) tasks[idx].predecessor = false;
	else alert('task out of range during delete operation');

	generateTaskMenu();
	generatePredecessorsTable();

	showLinkPredecsDocsCheckbox(existsPredecessor());

	recalculateDueDate('PredecessorsChanged');
}

function addPredecessor(idx) {
	if (tasks[idx]) tasks[idx].predecessor = true;
	else alert('task out of range during add operation');

	generateTaskMenu(idx);
	generatePredecessorsTable();

	showLinkPredecsDocsCheckbox(existsPredecessor());

	recalculateDueDate('PredecessorsChanged');
}

// creates content (HTML) of taskMenu (tasks allowed to be addeded to predecessors table)
function generateTaskMenu(clickedIdx) {
	// first get the list of tasks that can't be predecessors of current task, because
	// it will create a cycle within tasks dependencies if they become a predecessor
	var forbiddenTasks = getAllSuccessorIndicesOf(getIndexFromId(tasks, currentTaskID));

	var lastMenuItemIdx = -1;
	var taskName;
	for (var i = 0; i < tasks.length; i++) {
		// tasks not selected as predecessors and not in "forbidden list" are listed in taskMenu
		if (!tasks[i].predecessor && !arrayContainsValue(forbiddenTasks, i)) {
			taskName = trimStringToPx(tasks[i].name, taskMenuWidth);
			tasksMenu[i] = '<a href="javascript:addPredecessor(' + i + ');">' + WCXML.xmlencode(taskName) + '</a><br>';
			lastMenuItemIdx = i;
		} else {
			tasksMenu[i] = '';
		}
	}
	// handle the empty list of tasks in the taskMenu
	if (lastMenuItemIdx < 0) tasksMenu[0] = '&lt;' + langStrNoTaskToAdd + '&gt;';

	// update the contents of taskMenu immediately
	if ((typeof dropmenuobj != 'undefined') && dropmenuobj) populatemenu(tasksMenu);

	// start to hide list of tasks to be added among predecessors if
	// the clicked item is the last one -> mouse cursor gets out of
	// the menu this way...
	if ((typeof clickedIdx != 'undefined') && (clickedIdx > lastMenuItemIdx)) {
		// make the delay before menu hides longer in this case
		var origVal = disappeardelay;
		disappeardelay *= 2;
		delayhidemenu();
		disappeardelay = origVal;
	}
}

// creates content (HTML) of predecessors table 
function generatePredecessorsTable() {
	var table = $('<table class="table add-spacing-top">');

	var tbody = $('<tbody>');
	table.append($(generatePredTableHeader()));
	table.append(tbody);
	if (existsPredecessor()) {
		for (var i = 0; i < tasks.length; i++) {
			if (tasks[i] && tasks[i].predecessor) {
				tbody.append(generatePredTableRow(i));
			}
		}
	} else {
		// no predecessor specified - generate empty table
		tbody.append($('<tr><td align="center" colspan="4">&lt;' + WCXML.xmlencode(langStrNoPredecessor) + '&gt;</td></tr>'));
	}

	$("#" + predTableContainerID).empty().append(table);
}

function generatePredTableHeader() {
	var res = '<thead><tr>';
	res += '<th class="td-delete"/>';
	//  res += '<th style="white-space:nowrap;">#</th>';
	res += '<th class="td-nowrap">' + WCXML.xmlencode(langStrTaskName) + '</th>';
	res += '<th class="td-nowrap">' + WCXML.xmlencode(langStrAssignedTo) + '</th>';
	res += '<th class="td-nowrap td-date">' + WCXML.xmlencode(langStrDueDate) + '</th></tr></thead>';
	return res;
}

function generatePredTableRow(idx) {
	var row = tasks[idx];
	var res = $('<tr>');

	var s = '<td class="td-delete delete-predecessor">';
	if (isEditable && (taskStatusID == '00001_0000000001'))
		s += ' <i class="icon icon-normal icon-WCR_Filter-Remove12" value="' + row.id + '" onclick="deleteTaskPredecessor(' + idx + ');" title="' + WCXML.xmlencode(langStrDeletePredecessor) + '"></i>';
	s += '</td>';
	s += '<td class="task-predecessor-name">' + WCXML.xmlencode(row.name) + '</td>';
	s += '<td><nobr>' + WCXML.xmlencode(row.assignedTo) + '</nobr></td>';

	res.append($(s));

	var predDueDate = new Date(parseInt(row.dueDate));
	// convert given predecessor due date using current summer/winter time offset
	predDueDate = fixDateSummerWinterTime(predDueDate);

	var dueDateCell = $('<td class="td-nowrap td-date">');
	res.append(dueDateCell);
	if (row.dueDateAttributeID) {
		var value = {
			item_id: row.dueDateAttributeID,
			offset: row.dueDateOffset
		};
		if (row.dueDate !== 0) {
			value.time = predDueDate;
		}
		displayReadonlyPicker(dueDateCell, value);
	} else {
		var predDueDateText = WCHTML.getReadableDateAndTime(predDueDate);
		dueDateCell.text(predDueDateText);
	}
	return res;
}

function showLinkPredecsDocsCheckbox(show) {
	// initialize "can start task manually check box"
	var linkPredecsDocsElem = getElement('linkPredecsDocsWhenStarted');
	if (!linkPredecsDocsElem) return;
	if (show && isEditable) {
		linkPredecsDocsElem.disabled = false;
		linkPredecsDocsElem.checked = linkPredecsDocs;
	} else {
		linkPredecsDocsElem.disabled = true;
		linkPredecsDocsElem.checked = false;
	}
	//  linkPredecsDocs = linkPredecsDocsElem.checked;
}

/*************************************************************************/
/* END of methods related to task predecessors table */
/*************************************************************************/


function existsPredecessor() {
	for (var i = 0; i < tasks.length; i++) {
		if (tasks[i].predecessor) return true;
	}
	return false;
}

// find predecessor with the biggest dueDate value (the most in the future)
function getLastPredecessorIdx() {
	var longestPredIdx = null;
	var longestDueDate = -1;
	for (var i = 0; i < tasks.length; i++) {
		var task = tasks[i];
		if (!task.predecessor) continue;
		if (longestDueDate < task.dueDate) {
			longestDueDate = task.dueDate;
			longestPredIdx = i;
		}
	}
	return longestPredIdx;
}

function getLatestPredecessorDueDate() {
	var lastDueDate = getNow();
	// find predecessor with the latest dueDate value (the most in the future)
	var predIdx = getLastPredecessorIdx();
	if (predIdx != null) {
		var predDueDate = parseInt(tasks[predIdx].dueDate);
		if (predDueDate > 0) {
			lastDueDate.setTime(predDueDate);
		} else {
			lastDueDate = getNow();
		}
	}
	// convert given predecessor due date using current summer/winter time offset
	lastDueDate = fixDateSummerWinterTime(lastDueDate); //JANH *

	// return latest due date of all predecessors or current date and time if no predecessor found
	return lastDueDate;
}

function getNow() {
	var now = new Date();
	if (now.getMinutes() < 30) now.setMinutes(30);
	else {
		now.setHours(now.getHours() + 1);
		now.setMinutes(0);
	}

	return now;
}

function calculateStartDateFromDueDate(dueDate, leadDays, leadHours) {

	if ((actTaskType != null) && dueDate && !isNaN(leadDays) && !isNaN(leadHours)) {
		// subtract lead time from due date to get start date (note special handling of leadDays == 0 case)
		var stDate = (leadDays == 0) ? dueDate : dateSubWorkDays(dueDate, leadDays);
		stDate = dateSubWorkHoursMins(stDate, leadHours, 0);
		return stDate;
	} else return null;
}

function roundDateToLaterHalfHours(dd) {
	var mins = dd.getMinutes();
	if ((mins > 0) && (mins <= 30)) {
		// round to later half an hour
		dd.setMinutes(30);
	} else if (mins > 30) {
		// round to later hour
		dd.setMinutes(0);
		dd.setHours(dd.getHours() + 1);
	}
	dd.setSeconds(0);
	dd.setMilliseconds(0);

	return dd;
}

function roundDateToEarlierHalfHours(dd) {
	var mins = dd.getMinutes();
	if (mins < 30) {
		// round to the base hour (no minutes)
		dd.setMinutes(0);
	} else if ((mins >= 30) && (mins < 60)) {
		// round to later hour
		dd.setMinutes(30);
	}
	dd.setSeconds(0);
	dd.setMilliseconds(0);

	return dd;
}


// Note: Parameter "value" must be specified when mode == 'LeadTimeChanged' (it stands for 
// original lead time in hours).
function recalculateDueDate(mode, value) {
	
	callWhenDueDatePickerIsInDOM(function(){
		if (startDate == null) {
			// when startDate is not calculated/set yet do sonow from already initialized due date and lead time
			startDate = calculateStartDateFromDueDate(dueDate, leadDays, leadHours);
			if (startDate == null) {
				// startDate can't be calculated yet - just do the required "GUI work"
				updateStartDateGUI();
				return;
			}
		}
		
		if (typeof mode != 'undefined') {
			if (mode == 'PredecessorsChanged') {
				// continue only if in template or composing a new task or start option is "predecessor-sensitive"
				if (!isTemplate && !newTaskComposing && ((startTaskOptionIdx == 0) || (startTaskOptionIdx == 1))) return;
				
				var dueDateValue = getDueDatePickerValue();
				if (dueDateValue && dueDateValue.mode === "offset") {
					// when an attribute is set as due date, we don't change the start date
					return;
				} else {
					startDate = getLatestPredecessorDueDate();
				}
			} else if (mode == 'LeadTimeChanged') {
				if (isTemplate || (newTaskComposing && (dueDate == null))) {
					// special behaviour when lead time is changed in template or during composing of a new task
					startDate = getLatestPredecessorDueDate();
				} else {
					// lead time of an existing task in active project is edited 
					// when the due date is set by an attribute -> change the start date
					// when the due date is fixed -> recalculate the due date (happens below)
					var dueDateValue = getDueDatePickerValue();
					if (dueDateValue && dueDateValue.mode === "offset") {
						startDate = dateSubWorkDays(dueDate, leadDays);
						startDate = dateSubWorkHoursMins(startDate, leadHours, 0);
						startDate = roundDateToLaterHalfHours(startDate);
						updateStartDateGUI();
						return;
					}
				}
			} else if (mode == 'StartTaskOptionChanged') {
				// continue only if start option is "predecessor-sensitive"
				if ((startTaskOptionIdx == 0) || (startTaskOptionIdx == 1)) return;
				
				var dueDateValue = getDueDatePickerValue();
				if (dueDateValue && dueDateValue.mode === "offset") {
					// when an attribute is set as due date, we don't change the start date
					return;
				} else {
					startDate = getLatestPredecessorDueDate();
				}
			}
		}
		
		// add duration of current task to startDate to get the due date (counting only working days)
		dueDate = dateAddWorkDays(startDate, leadDays);
		dueDate = dateAddWorkHoursMins(dueDate, leadHours, 0);
		
		// ensure dueDate is rounded to nearest later half an hour
		dueDate = roundDateToLaterHalfHours(dueDate);
		
		dueDateRelatedChange = true;
		
		updateDueDateGUI(dueDate);
	});
}

// sets defined days and hours. if newDays/newHours are undefined values in form are not changed.
// if newDays/newHours are null the form values are reset to actual task type defaults, otherwise
// specified values are set...
function updateLeadTimeGUI(newDays, newHours) {
	var daysElem = getElement('lengthInput');
	var hoursElem = getElement('lengthTime');
	if (!daysElem || !hoursElem) return;

	if (actTaskType == null) {
		// no task type selected - reset and disable lead time GUI items
		daysElem.value = '';
		hoursElem.value = '';
		daysElem.disabled = true;
		hoursElem.disabled = true;
	} else {
		// a task type selected
		if (daysElem.disabled) {
			// first of all enable lead time GUI items
			daysElem.disabled = false;
			hoursElem.disabled = false;
		}
		if (typeof daysElem.value == 'undefined') {
			// read only lead time
			daysElem.innerHTML = leadDays + '';
			hoursElem.innerHTML = leadHours + '';
		} else {
			// editable lead time
			var taskType = taskTypes[actTaskType];
			if (typeof newDays == 'undefined') newDays = parseInt(daysElem.value);
			leadDays = ((newDays == null) || (isNaN(newDays))) ? taskType.days : newDays;
			if (typeof newHours == 'undefined') newHours = parseInt(hoursElem.value);
			leadHours = ((newHours == null) || (isNaN(newHours))) ? taskType.hours : newHours;

			daysElem.value = leadDays + '';
			hoursElem.value = leadHours + '';
		}
	}
}

function updateDueDateGUI(date) {
	var value = getDueDatePickerValue();
	if (value) {
		value.time = date ? date.getTime() : null;
		setDueDatePickerValue(value);
	}

	// update start date of current task info string
	updateStartDateGUI();
}

// Update start date of current task shown in the second and the fourth task start option string.
function updateStartDateGUI() {
	startDate = calculateStartDateFromDueDate(dueDate, leadDays, leadHours);

	// Wait until the relative date time picker code is loaded to show the "Reset Start Date" button and the UI for "Automatically at ..."
	require(["component/relativedatetimepicker/ResolvableRelativeDateTimePicker"], function (ResolvableRelativeDateTimePicker) {
		var dueDateValue = getDueDatePickerValue();
		if (isEditable && (taskStatusID == '00001_0000000001') && existsPredecessor() &&
			(startTaskOptionIdx != 0) && (startTaskOptionIdx != 1) && (startDate != null) &&
			(startDate.getTime() != getLatestPredecessorDueDate().getTime()) &&
			!(dueDateValue && dueDateValue.mode === "offset")) {
			$("#resetStartDateFormGroup").removeClass("hidden");

			// "start date reset" button HTML code
			var startDateResetButtonHtml = '<input type="button" name="startDateResetButton" class="btn btn-default"' +
				' onclick="recalculateDueDate(\'StartTaskOptionChanged\')" value="' + WCXML.xmlencode(langStrStartDateReset) + '" />';

			// show the "start date reset" button, but only if it makes sence
			setElementHtml('resetStartDateTD', startDateResetButtonHtml);
		} else {
			// hide the "start date reset" button
			setElementHtml('resetStartDateTD', '');
			$("#resetStartDateFormGroup").addClass("hidden");
		}

		// UI for "Automatically at ..."	
		if (dueDateValue && dueDateValue.mode === "offset" && (!dueDate || dueDate === 0)) {
			var millisecondsLeadTime = (leadDays * 24 + leadHours) * 3600000;
			dueDateValue.offset = dueDateValue.offset - millisecondsLeadTime; // The offset shown here is the combination of the offset to the attribute and the leadtime

			displayReadonlyPicker($("#taskStartDate1"), dueDateValue);
			displayReadonlyPicker($("#taskStartDate2"), dueDateValue);
		} else {
			var startDateStr = '???';
			if (startDate != null) startDateStr = WCHTML.getReadableDateAndTime(startDate);
			setElementHtml('taskStartDate1', '&nbsp;&nbsp;' + startDateStr + '&nbsp;&nbsp;');
			setElementHtml('taskStartDate2', '&nbsp;&nbsp;' + startDateStr + '&nbsp;&nbsp;');
		}
	});

}


/*****************************************************/
/* Saving and validation functions */
/*****************************************************/

function onSave(outputId) {

	// put the data of the due date picker in the form
	var div = $("#relativedatetimepicker_duedate");
	var picker = div.data("relativeDateTimePicker");
	if (picker) {
		var value = picker.getValue();
		if (value) {
			if (value.item_id) {
				$("#duedate_itemid").val(value.item_id);
				$("#duedate_offset").val(value.offset);
				$("#due_date").val(""); // if there is an attribute, do not store the value, it might be wrong
			} else {
				$("#due_date").val(value.time);
			}
		}
	}

	// get chosen predecessors' IDs from predecessors table
	var result = '';
	for (var i = 0; i < tasks.length; i++) {
		var task = tasks[i];
		// skip tasks that are not set as predecessors
		if (!task.predecessor) continue;
		// add space as a delimiter
		if (result.length > 0) result += ' ';
		result += task.id;
	}
	// save composed result into HTML input element to pass it to JSP control code
	var output = getElement(outputId);
	output.value = result;
}

/* 
 * This method returns 0 if all verified changes are valid. Otherwise it returns negative number 
 * that corresponds to the first found violation
 */
function validateMadeChanges() {
	// skip validation if there is nothing to validate
	if (!getElement('taskStatusID')) return 0;

	// verify validity of task status change
	if (!allowManualStart && (startTaskOptionIdx != 0)) {
		// if start task manually when "allow manual start" check box is not selected and start option
		// is not "Manually"
		if ((taskStatusID == '00001_0000000001') && (taskStatusID != getElement('taskStatusID').value)) {
			// task cannot be started manually
			return -1;
		}
	}
	return 0;
}

function validateDueDate() {
	// if task status is set to "Completed", any notification about overdue does not make sence
	if (getElement('taskStatusID').value == '00001_0000000003') return true;

	if (!dueDate) {
		var duedateValue = getDueDatePickerValue();
		if (!duedateValue || !duedateValue.item_id) {
			alert(langStrNoDueDateErr);
			return false;
		} else {
			// there is no actual due date value, but there is a due date attribute specified -> ok
			return true;
		}
	}

	// there was no action done by user that could effect due date - skip due dates checks
	if (!dueDateRelatedChange) return true;

	var ddts = dueDate.getTime(); // Due Date Time Stamp
	var curts = (new Date()).getTime(); // CURrent Time Stamp

	// due date should not be set to past
	if (curts > ddts) {
		return confirm(langStrDueDateInPastErr);
	}

	// validate start date based on predecessors due dates and current time
	if (startTaskOptionIdx != 0) {
		var startDate = dateSubWorkDays(dueDate, leadDays);
		startDate = dateSubWorkHoursMins(startDate, leadHours, 0);

		if ((startTaskOptionIdx == 2) || (startTaskOptionIdx == 3)) {
			// find predecessor with the biggest dueDate value (the most in the future)
			var predIdx = getLastPredecessorIdx();
			var longestDueDate = (predIdx == null) ? -1 : tasks[predIdx].dueDate;

			if (longestDueDate - 29 * 60000 > startDate.getTime()) {
				// (the "29*60000" miliseconds is a margin hiding insignificant differences...)
				return confirm(langStrTaskSequenceErr);
			}
		}

		if (((startTaskOptionIdx == 1) || (startTaskOptionIdx == 3)) && (curts > startDate.getTime())) {
			return confirm(langStrStartDateInPastErr);
		}
	}

	return true;
}

/**
 * Checks if a valid Group/Role Assignee was chosen if the Task is configured to be Restricted 
 */
function validateRestrictedAssignee() {
	//is the selected Task a 'restricted' one?
	var restrictAssignees = taskTypes[actTaskType].restrictAssignees;
	if (restrictAssignees === '1') {
		//does the parent restricted Task Type have a pre-selected Allowed Assignee reference?
		var ttPreselectedMA_ID = taskTypes[actTaskType].preselectedMainAssigneeID;
		if (!ttPreselectedMA_ID || ttPreselectedMA_ID.length === 0) {
			//check the 'Assigned to' dropdown value > should be a Role/Group > the option value has to begin with 'G'
			if (!($('#taskOwner').val().startsWith("G"))) {
				return false;
			}
		}
	}
	
	return true;
}


/* ========================== other functions related to wctasks ========================== */


function sortNumber(a, b) {
	return a - b;
}

function generateDependsOnList(elemId, idList, translTable) {
	var str = '';
	var indices = new Array();

	// translate list of task IDs (idList) to list of row numbers in table of tasks
	for (var v = 0; v < idList.length; v++) {
		var id = idList[v];
		for (var i = 0; i < translTable.length; i++) {
			if (translTable[i].id == id) indices[indices.length] = translTable[i].value;
		}
	}
	// sort the list of predecessor indices (row numbers in table of tasks)
	indices.sort(sortNumber);

	if (indices.length > 0) {
		for (var i = 0; i < indices.length - 1; i++) {
			str += indices[i] + ', ';
		}
		str += indices[indices.length - 1];
	}

	setElementHtml(elemId, str);
}
